﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace TestProject
{
    /// <summary>
    /// 비트맵 래퍼
    /// </summary>
    public class BitmapWrapper : IDisposable, IEnumerable<Point>
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 소스 비트맵
        /// </summary>
        private Bitmap sourceBitmap;

        /// <summary>
        /// 소스 비트맵 데이터
        /// </summary>
        private BitmapData sourceBitmapData;

        /// <summary>
        /// 소스 바이트 배열
        /// </summary>
        private byte[] sourceByteArray;

        /// <summary>
        /// 복사 바이트 배열
        /// </summary>
        private byte[] copyByteArray;

        /// <summary>
        /// 스트라이드
        /// </summary>
        private int stride;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 비트맵 너비 - BitmapWidth

        /// <summary>
        /// 비트맵 너비
        /// </summary>
        public int BitmapWidth { get; private set; }

        #endregion
        #region 비트맵 높이 - BitmapHeight

        /// <summary>
        /// 비트맵 높이
        /// </summary>
        public int BitmapHeight { get; private set; }

        #endregion
        #region 디폴트 색상 - DefaultColor

        /// <summary>
        /// 디폴트 색상
        /// </summary>
        public Color DefaultColor { get; set; }

        #endregion

        #region 인덱서 - this[x, y]

        /// <summary>
        /// 인덱서
        /// </summary>
        /// <param name="x">X 좌표</param>
        /// <param name="y">Y 좌표</param>
        /// <returns>색상</returns>
        public Color this[int x, int y]
        {
            get
            {
                int i = GetIndex(x, y);

                return i < 0 ? DefaultColor : Color.FromArgb(sourceByteArray[i + 3], sourceByteArray[i + 2], sourceByteArray[i + 1], sourceByteArray[i]);
            }
            set
            {
                int i = GetIndex(x, y);

                if(i >= 0)
                {
                    copyByteArray[i    ] = value.B;
                    copyByteArray[i + 1] = value.G;
                    copyByteArray[i + 2] = value.R;
                    copyByteArray[i + 3] = value.A;
                };
            }
        }

        #endregion
        #region 인덱서 - this[point]

        /// <summary>
        /// 인덱서
        /// </summary>
        /// <param name="point">포인트</param>
        /// <returns>색상</returns>
        public Color this[Point point]
        {
            get
            {
                return this[point.X, point.Y];
            }
            set
            {
                this[point.X, point.Y] = value;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - BitmapWrapper(sourceBitmap, copySourceBitmap)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="sourceBitmap">소스 비트맵</param>
        /// <param name="copySourceBitmap">소스 비트맵 복사 여부</param>
        public BitmapWrapper(Bitmap sourceBitmap, bool copySourceBitmap = false)
        {
            BitmapWidth  = sourceBitmap.Width;
            BitmapHeight = sourceBitmap.Height;

            this.sourceBitmap = sourceBitmap;

            this.sourceBitmapData = sourceBitmap.LockBits
            (
                new Rectangle(0, 0, sourceBitmap.Width, sourceBitmap.Height),
                ImageLockMode.ReadWrite,
                PixelFormat.Format32bppArgb
            );

            this.stride = this.sourceBitmapData.Stride;

            this.sourceByteArray = new byte[this.stride * BitmapHeight];

            Marshal.Copy(this.sourceBitmapData.Scan0, this.sourceByteArray, 0, this.sourceByteArray.Length);

            this.copyByteArray = copySourceBitmap ? (byte[])sourceByteArray.Clone() : new byte[this.stride * BitmapHeight];
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 픽셀 설정하기 - SetPixel(point, red, green, blue)

        /// <summary>
        /// 픽셀 설정하기
        /// </summary>
        /// <param name="point">포인트</param>
        /// <param name="red">빨간색</param>
        /// <param name="green">녹색</param>
        /// <param name="blue">파란색</param>
        public void SetPixel(Point point, double red, double green, double blue)
        {
            if(red < 0)
            {
                red = 0;
            }

            if(red >= 256)
            {
                red = 255;
            }

            if(green < 0)
            {
                green = 0;
            }

            if(green >= 256)
            {
                green = 255;
            }

            if(blue < 0)
            {
                blue = 0;
            }

            if(blue >= 256)
            {
                blue = 255;
            }

            this[point.X, point.Y] = Color.FromArgb((int)red, (int)green, (int)blue);
        }

        #endregion
        #region 교환하기 - Swap()

        /// <summary>
        /// 교환하기
        /// </summary>
        public void Swap()
        {
            byte[] temporaryByteArray = sourceByteArray;

            sourceByteArray = copyByteArray;

            copyByteArray = temporaryByteArray;
        }

        #endregion
        #region 열거자 구하기 - GetEnumerator()

        /// <summary>
        /// 열거자 구하기
        /// </summary>
        /// <returns>포인트 열거자</returns>
        public IEnumerator<Point> GetEnumerator()
        {
            for(int y = 0; y < BitmapHeight; y++)
            {
                for(int x = 0; x < BitmapWidth; x++)
                {
                    yield return new Point(x, y);
                }
            }
        }

        #endregion
        #region (IEnumerable) 열거자 구하기 - GetEnumerator()

        /// <summary>
        /// 열거자 구하기
        /// </summary>
        /// <returns></returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion
        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            Marshal.Copy(copyByteArray, 0, sourceBitmapData.Scan0, copyByteArray.Length);

            sourceBitmap.UnlockBits(sourceBitmapData);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 인덱스 구하기 - GetIndex(x, y)

        /// <summary>
        /// 인덱스 구하기
        /// </summary>
        /// <param name="x">X 좌표</param>
        /// <param name="y">Y 좌표</param>
        /// <returns></returns>
        private int GetIndex(int x, int y)
        {
            return (x < 0 || x >= BitmapWidth || y < 0 || y >= BitmapHeight) ? -1 : x * 4 + y * stride;
        }

        #endregion
    }
}