﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 확대/축소 픽처 박스
    /// </summary>
    public class ZoomPictureBox : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 표시 중심점 변경시 이벤트 - VisibleCenterPointChanged

        /// <summary>
        /// 표시 중심점 변경시 이벤트
        /// </summary>
        public event EventHandler VisibleCenterPointChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 마우스 상태 - MouseState

        /// <summary>
        /// 마우스 상태
        /// </summary>
        private enum MouseState
        {
            /// <summary>
            /// NONE
            /// </summary>
            None,
            
            /// <summary>
            /// DRAG
            /// </summary>
            Drag
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 소스 비트맵
        /// </summary>
        private Bitmap sourceBitmap;

        /// <summary>
        /// 표시 중심점
        /// </summary>
        private PointF visibleCenterPoint;

        /// <summary>
        /// 확대/축소
        /// </summary>
        private float zoom = 1f;

        /// <summary>
        /// 마우스 상태
        /// </summary>
        private MouseState mouseState;

        /// <summary>
        /// 드래그 시작 포인트
        /// </summary>
        private Point dragStartPoint;

        /// <summary>
        /// 드래그 시작 표시 중심점
        /// </summary>
        private PointF dragStartVisibleCenterPoint;

        /// <summary>
        /// 소스 비트맵 너비
        /// </summary>
        private int sourceBitmapWidth;

        /// <summary>
        /// 소스 비트맵 높이
        /// </summary>
        private int sourceBitmapHeight;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 축소/확대 델타 - ZoomDelta

        /// <summary>
        /// 축소/확대 델타
        /// </summary>
        [DefaultValue(0.1f)]
        public float ZoomDelta { get; set; }

        #endregion
        #region 사용자 드래그 허용 여부 - AllowUserDrag

        /// <summary>
        /// 사용자 드래그 허용 여부
        /// </summary>
        [DefaultValue(true)]
        public bool AllowUserDrag { get; set; }

        #endregion
        #region 사용자 확대/축소 허용 여부 - AllowUserZoom

        /// <summary>
        /// 사용자 확대/축소 허용 여부
        /// </summary>
        [DefaultValue(true)]
        public bool AllowUserZoom { get; set; }

        #endregion
        #region 보간 모드 - InterpolationMode

        /// <summary>
        /// 보간 모드
        /// </summary>
        public InterpolationMode InterpolationMode { get; set; }

        #endregion
        #region 픽셀 오프셋 모드 - PixelOffsetMode

        /// <summary>
        /// 픽셀 오프셋 모드
        /// </summary>
        public PixelOffsetMode PixelOffsetMode { get; set; }

        #endregion
        #region 소스 비트맵 - SourceBitmap

        /// <summary>
        /// 소스 비트맵
        /// </summary>
        [DefaultValue(null)]
        public Bitmap SourceBitmap
        {
            get
            {
                return this.sourceBitmap;
            }
            set
            {
                this.sourceBitmap = value;

                if(value == null)
                {
                    this.sourceBitmapWidth  = 0;
                    this.sourceBitmapHeight = 0;

                    VisibleCenterPoint = new PointF(0, 0);
                }
                else
                {
                    this.sourceBitmapWidth  = value.Width;
                    this.sourceBitmapHeight = value.Height;

                    VisibleCenterPoint = new PointF(this.sourceBitmapWidth / 2f, this.sourceBitmapHeight / 2f);
                }

                Invalidate();
            }
        }

        #endregion
        #region 확대/축소 - Zoom

        /// <summary>
        /// 확대/축소
        /// </summary>
        [DefaultValue(1f)]
        public float Zoom
        {
            get
            {
                return this.zoom;
            }

            set
            {
                if(Math.Abs(value) <= float.Epsilon)
                {
                    throw new Exception("Zoom must be more then 0");
                }

                this.zoom = value;

                Invalidate();
            }
        }

        #endregion
        #region 표시 중심 포인트 - VisibleCenterPoint

        /// <summary>
        /// 표시 중심 포인트
        /// </summary>
        public PointF VisibleCenterPoint
        {
            get
            {
                return this.visibleCenterPoint;
            }

            set
            {
                this.visibleCenterPoint = value;

                OnVisibleCenterPointChanged();
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - ZoomPictureBox()

        /// <summary>
        /// 생성자
        /// </summary>
        public ZoomPictureBox()
        {
            SetStyle
            (
                ControlStyles.AllPaintingInWmPaint | ControlStyles.UserPaint | ControlStyles.OptimizedDoubleBuffer | ControlStyles.ResizeRedraw,
                true
            );

            ZoomDelta = 0.1f;

            AllowUserDrag = true;

            AllowUserZoom = true;

            InterpolationMode = InterpolationMode.Bicubic;
            PixelOffsetMode   = PixelOffsetMode.HighQuality;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 표시 중심 포인트 변경시 처리하기 - OnVisibleCenterPointChanged()

        /// <summary>
        /// 표시 중심 포인트 변경시 처리하기
        /// </summary>
        public virtual void OnVisibleCenterPointChanged()
        {
            VisibleCenterPointChanged?.Invoke(this, EventArgs.Empty);

            Invalidate();
        }

        #endregion
        #region 비트맵 업데이트 하기 - UpdateBitmap(sourceBitmap)

        /// <summary>
        /// 비트맵 업데이트 하기
        /// </summary>
        /// <param name="sourceBitmap">소스 비트맵</param>
        public void UpdateBitmap(Bitmap sourceBitmap)
        {
            this.sourceBitmap = sourceBitmap;

            if(sourceBitmap != null)
            {
                this.sourceBitmapWidth  = sourceBitmap.Width;
                this.sourceBitmapHeight = sourceBitmap.Height;
            }

            Invalidate();
        }

        #endregion

        #region 축소하기 - DecreaseZoom()

        /// <summary>
        /// 축소하기
        /// </summary>
        public void DecreaseZoom()
        {
            Zoom = (float)Math.Exp(Math.Log(this.zoom) - ZoomDelta);
        }

        #endregion
        #region 확대하기 - IncreazeZoom()

        /// <summary>
        /// 확대하기
        /// </summary>
        public void IncreazeZoom()
        {
            Zoom = (float)Math.Exp(Math.Log(this.zoom) + ZoomDelta);
        }

        #endregion

        #region 비트맵 포인트 구하기 - GetBitmapPoint(clientPoint)

        /// <summary>
        /// 비트맵 포인트 구하기
        /// </summary>
        /// <param name="clientPoint">클라이언트 포인트</param>
        /// <returns>비트맵 포인트</returns>
        public Point GetBitmapPoint(Point clientPoint)
        {
            float deltaX = (clientPoint.X - ClientSize.Width  / 2f) / Zoom + this.visibleCenterPoint.X;
            float deltaY = (clientPoint.Y - ClientSize.Height / 2f) / Zoom + this.visibleCenterPoint.Y;

            return new Point((int)deltaX, (int)deltaY);
        }

        #endregion
        #region 클라이언트 포인트 구하기 - GetClientPoint(bitmapPoint)

        /// <summary>
        /// 클라이언트 포인트 구하기
        /// </summary>
        /// <param name="bitmapPoint">비트맵 포인트</param>
        /// <returns>클라이언트 포인트</returns>
        public Point GetClientPoint(Point bitmapPoint)
        {
            float deltaX = (bitmapPoint.X - visibleCenterPoint.X) * Zoom + ClientSize.Width  / 2f;
            float deltaY = (bitmapPoint.Y - visibleCenterPoint.Y) * Zoom + ClientSize.Height / 2f;

            return new Point((int)deltaX, (int)deltaY);
        }

        #endregion

        #region 스크린샷 구하기 - GetScreenshot()

        /// <summary>
        /// 스크린샷 구하기
        /// </summary>
        /// <returns>스크린샷</returns>
        public Image GetScreenshot()
        {
            Image targetImage = new Bitmap(ClientSize.Width, ClientSize.Height);

            using(Graphics targetGraphics = Graphics.FromImage(targetImage))
            {
                OnPaint(new PaintEventArgs(targetGraphics, ClientRectangle));
            }

            return targetImage;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 마우스 휠 처리하기 - OnMouseWheel(e)

        /// <summary>
        /// 마우스 휠 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseWheel(MouseEventArgs e)
        {
            base.OnMouseWheel(e);

            if(!AllowUserZoom)
            {
                return;
            }

            if(e.Delta > 0)
            {
                IncreazeZoom();
            }

            if(e.Delta < 0)
            {
                DecreaseZoom();
            }
        }

        #endregion
        #region 마우스 DOWN 처리하기 - OnMouseDown(e)

        /// <summary>
        /// 마우스 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if(AllowUserDrag)
            {
                if(e.Button == MouseButtons.Left)
                {
                    Cursor = Cursors.SizeAll;

                    this.mouseState = MouseState.Drag;
                }
            }

            this.dragStartPoint = e.Location;

            this.dragStartVisibleCenterPoint = VisibleCenterPoint;
        }

        #endregion
        #region 마우스 이동시 처리하기 - OnMouseMove(e)

        /// <summary>
        /// 마우스 이동시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if(mouseState == MouseState.Drag)
            {
                int deltaX = e.Location.X - this.dragStartPoint.X;
                int deltaY = e.Location.Y - this.dragStartPoint.Y;

                VisibleCenterPoint = new PointF
                (
                    this.dragStartVisibleCenterPoint.X - deltaX / this.zoom,
                    this.dragStartVisibleCenterPoint.Y - deltaY / this.zoom
                );
            }
        }

        #endregion
        #region 마우스 UP 처리하기 - OnMouseUp(e)

        /// <summary>
        /// 마우스 UP 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            Cursor = Cursors.Default;

            this.mouseState = MouseState.None;

            Invalidate();
        }

        #endregion

        #region 페인트시 처리하기 - OnPaint(e)

        /// <summary>
        /// 페인트시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnPaint(PaintEventArgs e)
        {
            if(this.sourceBitmap == null)
            {
                return;
            }

            e.Graphics.ResetTransform();

            e.Graphics.InterpolationMode = InterpolationMode;
            e.Graphics.PixelOffsetMode   = PixelOffsetMode;

            if(this.mouseState == MouseState.Drag || Zoom < 1f)
            {
                e.Graphics.InterpolationMode = InterpolationMode.NearestNeighbor;
            }

            float deltaX = -this.visibleCenterPoint.X * Zoom + ClientSize.Width  / 2f;
            float deltaY = -this.visibleCenterPoint.Y * Zoom + ClientSize.Height / 2f;

            e.Graphics.DrawImage
            (
                this.sourceBitmap,
                deltaX,
                deltaY,
                this.sourceBitmap.Width  * Zoom,
                this.sourceBitmap.Height * Zoom
            );

            base.OnPaint(e);
        }

        #endregion
    }
}